# Copyright (C) 2006-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import errno
import os
import codecs
import socket
import logging
import xml.dom.minidom

from agent.api.network.base import Network
from agent.extension import OVSDiscoverExtension
from agent.lib import linux_netutils, filelock
from agent.lib.config import get_config
from agent.target.host import run_cmd


class LinuxNetwork(Network):
    """
    Class for network management.
    """

    def __init__(self):
        config = get_config()
        try:
            virtualnetwork = config.get('network', 'virtualnetwork').lower()
        except Exception:
            virtualnetwork = 'linuxbridge'
        if virtualnetwork == 'openvswitch':
            self.script_dir = '/etc/xen/scripts/openvswitch'
        else:
            self.script_dir = '/etc/xen/scripts/linuxbridge'
        Network.__init__(self)

    def get_services(self, version=None):
        return {
            'configure_virtual_ip': self.configure_virtual_ip,
            'deconfigure_virtual_ip': self.deconfigure_virtual_ip,
            'ovs_br_config': self.ovs_br_config,
            'ovs_async_bridge': self.ovs_async_bridge,
            'ovs_vlan_bridge': self.ovs_vlan_bridge,
            'ovs_bond_config': self.ovs_bond_config,
            'ovs_async_bond': self.ovs_async_bond,
            'ovs_bond_op': self.ovs_bond_op,
            'ovs_bond_mode': self.ovs_bond_mode,
            'ovs_local_config': self.ovs_local_config,
            'ovs_vlan_config': self.ovs_vlan_config,
            'ovs_ip_config': self.ovs_ip_config,
            'ovs_if_meta': self.ovs_if_meta,
            'ovs_change_mtu': self.ovs_change_mtu,
            'discover_network': self.discover_network,
            'ovs_ovca_network': self.ovs_ovca_network,
        }

    def ovs_br_config(self, action, br_name, net_dev):
        """
        Configure a Standard Bridge.

        @param action (start | stop)
              start: Creates the bridge, Copies the IP and MAC addresses from netdev to bridge,
                     enslaves net_dev to bridge, and saves bridge config files in
                     /etc/sysconfig/network-scripts
              stop:  Removes net_dev from the bridge,transfers addresses, routes from bridge to
                     net_dev, deletes the bridge, and revomes bridge config files from
                     /etc/sysconfig/network-scripts

        @param br_name       The bridge name to add
        @param net_dev       The physical interface to add to the bridge

        @return    If successful, returns the names of the bridge and it's physical interface.
                   Raises an exception on failure
        Restriction:
        -net_dev  must be physical, or bond
        -net_dev  must not be member of a bridge, or slave to a bond
        """
        if action != "start" and action != "stop":
            raise Exception ("action %s is not supported" % action)
        if br_name.startswith("bond") or br_name.startswith("eth"):
            raise Exception ("bridge name can not be prefixed with (bond or eth) %s" % br_name)
        cmd = ["%s/ovs-network-bridge" % self.script_dir, action, "bridge=%s" % br_name, "netdev=%s" % net_dev]
        run_cmd(cmd)
        logging.info("ovs-network-bridge: success. %s bridge %s on interface %s" % (action, br_name, net_dev))
        return "%s bridge=%s netdev=%s" % (action, br_name, net_dev)

    def ovs_async_bridge(self, action, br_name, net_dev):
        """
        Configure a Standard Bridge.This command to be used in async mode.

        @param action (start | stop)
              start: Creates the bridge, Copies the IP and MAC addresses from netdev to bridge,
                     enslaves net_dev to bridge, and saves bridge config files in
                     /etc/sysconfig/network-scripts
              stop:  Removes net_dev from the bridge,transfers addresses, routes from bridge to
                     net_dev, deletes the bridge, and revomes bridge config files from
                     /etc/sysconfig/network-scripts

        @param br_name       The bridge name to add
        @param net_dev       The physical interface to add to the bridge

        @return    If successful, returns the names of the bridge and it's physical interface.
                   Raises an exception on failure
        Restriction:
        -net_dev  must be physical, or bond
        -net_dev  must not be member of a bridge, or slave to a bond
        """
        if action != "start" and action != "stop":
            raise Exception ("action %s is not supported" % action)
        if br_name.startswith("bond") or br_name.startswith("eth"):
            raise Exception ("bridge name can not be prefixed with (bond or eth) %s" % br_name)
        cmd = ["%s/ovs-network-bridge" % self.script_dir, action, "bridge=%s" % br_name, "netdev=%s" % net_dev]
        run_cmd(cmd)

    def ovs_vlan_bridge(self, action, br_name, net_dev, vlan_id):
        """
        Configure a VLAN Bridge.

        @param action (start | stop)
              start: Creates the bridge, creats VLAN on net_dev,enslaves
                     the VLAN to the bridge, and saves VLAN bridge config
                     files in /etc/sysconfig/network-scripts
              stop:  Removes the VLAN from the bridge, removes the VLAN,
                     deletes the bridge, and removes VLAN bridge config files
                     from /etc/sysconfig/network-scripts

        @param br_name       The bridge name to add
        @param net_dev       The physical interface on which to create a VLAN.
        @param vlan_id       VLAN ID (1-4095). VLAN ID of 1 is the untagged VLAN.

        @return       If successful, returns the names of the bridge and it's VLAN interface
                      Raises an exception on failure
        Restriction:
        -net_dev  must be physical, or bond
        -net_dev  must not be member of a bridge, or slave to a bond
        -VLAN ID must not exist on the same interface
        """
        if action != "start" and action != "stop":
            raise Exception ("action %s is not supported" % action)
        if br_name.startswith("bond") or br_name.startswith("eth"):
            raise Exception ("bridge name can not be prefixed with (bond or eth) %s" % br_name)
        cmd = ["%s/ovs-vlan-bridge" % self.script_dir, action, "bridge=%s" % br_name,
               "vlanid=%d" % int(vlan_id), "netdev=%s" % net_dev]
        run_cmd(cmd)
        logging.info("ovs-vlan-bridge: success. %s bridge %s on interface %s vlan %d" % (
                     action, br_name, net_dev, int(vlan_id)))
        return "%s bridge=%s netdev=%s vlan %d" % (action, br_name, net_dev, int(vlan_id))

    def ovs_bond_config(self, action, bond, *args):
        """
        Configure Bonding.

        @param action (start | stop)
              start: Bond two physical interfaces to one logical interface.
                     (eth0,eth1 --> bond0), and saves bond config files in
                     /etc/sysconfig/network-scripts
              stop:  Removes channel bond relation, removes bond config files from
                     /etc/sysconfig/network-scripts

        @param bond   One of the logical channel bonds (bond0, bond1 ...etc)
        @param args   Is a list of the following arguments: (Not required when removing a bond)

              bondmode   Current supported bonding modes (1 = active-backup, 4 = 802.3ad
                         6 = balance-alb).
              primary    Primary physical interface (eth0,eth1 ..etc).
              backup     Backup physical interface  (eth0,eth1 ..etc)
                         Primary/backup are only valid for active-backup mode,
                         they are used as a place holder when configuring the other two modes.

        @return       If successful, returns the names of the physical interfaces and the bond
                      Raises an exception on failure
        Restriction:
        -bond     must be one of the logical channel bonds (bond0, bond1 ...etc)
        -primary  must be physical, and not be member of a bridge, or slave to a bond
        -backup   must be physical, and not be member of a bridge, or slave to a bond

        """
        if action != "start" and action != "stop":
            raise Exception ("action %s is not supported" % action)
        if action == "start":
            if len(args) < 3:
                raise Exception ("ovs-bond-if failed: bond mode required")
            mode = args[0]
            primary = args[1]
            backup = args[2]
            cmd = ["%s/ovs-bond-if" % self.script_dir, action, "bondm=%s" % bond,
                   "bondmode=%d" % int(mode), "primary=%s" % primary, "backup=%s" % backup]
        else:
            cmd = ["%s/ovs-bond-if" % self.script_dir, action, "bondm=%s" % bond]
        run_cmd(cmd)
        if action == "start":
            logging.info("ovs-bond-if: success. %s bonding interfaces %s %s" % (action, primary, backup))
            return "%s bonding inetrfaces %s %s" % (action, primary, backup)
        else:
            logging.info("ovs-bond-if: success. removed %s" % (bond))
            return "removed bond %s" % (bond)

    def ovs_async_bond(self, action, bond, *args):
        """
        Configure Bonding. This command to be used in async mode.

        @param action (start | stop)
              start: Bond two physical interfaces to one logical interface.
                     (eth0,eth1 --> bond0), and saves bond config files in
                     /etc/sysconfig/network-scripts
              stop:  Removes channel bond relation, removes bond config files from
                     /etc/sysconfig/network-scripts

        @param bond   One of the logical channel bonds (bond0, bond1 ...etc)
        @param args   Is a list of the following arguments: (Not required when removing a bond)

              bondmode   Current supported bonding modes (1 = active-backup, 4 = 802.3ad
                         6 = balance-alb).
              primary    Primary physical interface (eth0,eth1 ..etc).
              backup     Backup physical interface  (eth0,eth1 ..etc)
                         Primary/backup are only valid for active-backup mode,
                         they are used as a place holder when configuring the other two modes.

        @return       If successful, returns the names of the physical interfaces and the bond
                      Raises an exception on failure
        Restriction:
        -bond     must be one of the logical channel bonds (bond0, bond1 ...etc)
        -primary  must be physical, and not be member of a bridge, or slave to a bond
        -backup   must be physical, and not be member of a bridge, or slave to a bond

        """
        if action != "start" and action != "stop":
            raise Exception ("action %s is not supported" % action)
        if action == "start":
            if len(args) < 3:
                raise Exception ("ovs-bond-if failed: bond mode required")
            mode = args[0]
            primary = args[1]
            backup = args[2]
            cmd = ["%s/ovs-bond-if" % self.script_dir, action, "bondm=%s" % bond,
                   "bondmode=%d" % int(mode), "primary=%s" % primary, "backup=%s" % backup]
        else:
            cmd = ["%s/ovs-bond-if" % self.script_dir, action, "bondm=%s" % bond]
        run_cmd(cmd)

    def ovs_bond_op(self, action, bond, backup):
        """
        Configure Bonding.

        @param action (add | remove)
              add: Adds a physical interfaces to an existing bond.
                     (eth0,eth1 --> bond0), and saves bond config files in
                     /etc/sysconfig/network-scripts
              remove:  Removes an interfacse from an existing bond and restore the
                       interface config file to /etc/sysconfig/network-scripts

        @param bond   One of the logical channel bonds (bond0, bond1 ...etc)
        @param backup  Physical interface to add/remove (eth0, eth1 ...etc)

        @return       If successful, returns the names of the physical interfaces and the bond
                      Raises an exception on failure
        Restriction:
        -bond     must be one of the logical channel bonds (bond0, bond1 ...etc)
        -backup   must be physical, and not be member of a bridge, or slave to a bond

        """
        if action != "add" and action != "remove":
            raise Exception ("action %s is not supported" % action)
        cmd = ["%s/ovs-bond-op" % self.script_dir, action, "bondm=%s" % bond, "backup=%s" % backup]
        run_cmd(cmd)
        if action == "add":
            logging.info("ovs-bond-op: success. added interface %s to %s" % (backup, bond))
            return "added inetrface %s to %s" % (backup, bond)
        else:
            logging.info("ovs-bond-op: success. removed interface %s from %s" % (backup, bond))
            return "removed interface %s form %s" % (backup, bond)

    def ovs_bond_mode(self, bond, mode):
        """
        Change Bond mode.

        @param bond   One of the logical channel bonds (bond0, bond1 ...etc)
        @mode         Current supported bonding modes (1 = active-backup, 4 = 802.3ad
                      6 = balance-alb).
        @return       If successful, returns bond's names and its new mode
                      Raises an exception on failure
        Restriction:
        -bond     must be one of the logical channel bonds (bond0, bond1 ...etc)
        """
        if not bond:
            raise Exception ("bond name is required")
        if not mode:
            raise Exception ("bond mode (1,4,6) is required")
        cmd = ["%s/ovs-bond-mode" % self.script_dir, "bondm=%s" % bond, "bondmode=%d" % int(mode)]
        run_cmd(cmd)
        logging.info("ovs-bond-mode: success. %s mode changed to %s" % (bond, mode))
        return "%s mode changed to %s" % (bond, mode)

    def ovs_local_config(self, action, br_name ):
        """
        Configure a Local Bridge ((NIC-less bridge)

        @param action (start | stop)
             start: Creates local bridge without a physical interface, and saves bridge config
                    file in /etc/sysconfig/network-scripts
             stop:  Deletes local bridge, removes bridge config file from
                    /etc/sysconfig/network-scripts

        @param br_name       The bridge name to add

        @return       If successful, returns the name of the bridge
                      Raises an exception on failure
        """
        if action != "start" and action != "stop":
            raise Exception ("action %s is not supported" % action)
        cmd = ["%s/ovs-local-bridge" % self.script_dir, action, "bridge=%s" % br_name]
        run_cmd(cmd)
        logging.info("ovs-local-bridge: success. %s local bridge %s" % (action, br_name))
        return "%s local bridge %s" % (action, br_name)

    def ovs_vlan_config(self, action, interface, vlanid ):
        """
        Creates a VLAN interface on a physical, or a bond interface.

        @param  action (add|remove)
              add:    Creates a VLAN on an interface,saves the VLAN config file in
                      /etc/sysconfig/network-scripts
              remove: Removes a VLAN from an interfacei,removes its config file from
                      /etc/sysconfig/network-scripts

        @param   interface    The interface on which to create a VLAN
        @param   vlanid       VLAN ID (2-4095)

        @return  If successful, returns the interface, and VLAN created
                 Raises an exception on failure

        Restriction:
        -Interface  must be physical, or bond
        -Interface must not be member of a bridge, or slave to a bond
        -VLAN ID must not exist on the same interface
        """

        if action != "add" and action != "remove":
            raise Exception ("action %s is not supported" % action)
        cmd = ["%s/ovs-vlan-config" % self.script_dir, action, interface, "%d" % int(vlanid)]
        run_cmd(cmd)
        logging.info("ovs-vlan-config: success. %s VLAN %d to %s" % (action, int(vlanid), interface))
        return "%s VLAN %d interface %s" % (action, int(vlanid), interface)


    def ovs_ip_config(self, interface, optype, *args):
        """
        Assigns/flushes IP, netmask address to a physical,VLAN, and bond interfaces.

        @param  interface    The interface on which to assign
        @param  optype (static|dynamic|flush)
            static:  Assigns the given IP, and netmask to the interface, and
                         saves the config file to /etc/sysconfig/network-scripts.
            dynamic: Flushes current address, and creats and save the config
                         file to /etc/sysconfig/network-scripts, (BOOTPROTO=dhcp)
            flush:   flushes the interface address,routes, removes the current
                         config file from /etc/sysconfig/network-scripts.
                         Creats a new one with BOOTPROTO=static

        @args   Required for the static option, otherwise it is ignored
            IP address:  IPv4 address in decimal notation (101.230.112)
            netmask:     Standard netmask in a decimal notation,NOT CIDR.
                             example(255.255.255.0)

        @return  If successful, returns the interface, and addresses added/flushed
                 Raises an exception on failure

        Restriction:
        -Interface must be physical, VLAN, or a Bond
        -Interface must not be a bridge port, or slave to a bond
        -Addresses must be valid in a decimal notation
        """
        if optype != "static" and optype != "dynamic" and optype != "flush":
            raise Exception ("operation type %s is not supported" % optype )
        if optype == "static":
            if len(args) < 2:
                raise Exception ("ovs-ip-config faild: IP Adress and netmask are required")
            ipaddr = args[0]
            netmask = args[1]

            #Validate IP
            try:
                socket.inet_aton(ipaddr)
                seprator = ipaddr.split(".")
            except:
                raise Exception ("IP %s is not valid." % ipaddr)
            if len(seprator) != 4:
                raise Exception ("ovs-ip-config faild: IP %s is not valid." % ipaddr)

            #Validate netmask
            seprator = netmask.split(".")
            if len(seprator) != 4:
                raise Exception ("ovs-ip-config faild: netmask %s is not legal valid." % netmask )

            cmd = ["%s/ovs-ip-config" % self.script_dir, interface, optype, ipaddr, netmask]
        else:
            cmd = ["%s/ovs-ip-config" % self.script_dir, interface, optype]

        run_cmd(cmd)
        if optype == "static":
            logging.info("ovs-ip-config: success: interface %s  IP %s mask %s" % (interface, ipaddr, netmask))
            return "interface %s  IP %s mask %s" % (interface, ipaddr, netmask)
        else:
            logging.info("ovs-ip-config: success: interface %s %s" % (interface, optype))
            return "interface %s %s" % (interface, optype)

    def ovs_if_meta(self, interface, data):
        """
        This function creates meta data file meta-interface, and write the string
        (METADATA=data) to it.  This string is used by the manager to identify
        networks that interfaces belong to. Dom0 does not make used of this string,
        it just saves it and returns it during running, saved network discovery.

         - If an interface already has a meta data string, then it gets
           replace by the new one
         - An empty meta data string, indicates to remove the existing string
           (remove the meta-interface) file

        @param  interface    physical,VLAN, bond ...etc interfaces
        @param  data         meta data to save

        @return  If successful, returns the interface, and meta data
                 Raises an exception on failure

        Restriction:
        - data string that starts with leading spaces will be rejected
        """

        if data.startswith(" "):
            raise Exception("ovs_set_metadata: meta data must not start with leading white space")

        cfg_file = "/etc/sysconfig/network-scripts/meta-%s" % interface
        if data == '': #remove meta data file
            try:
                os.unlink(cfg_file)
                logging.info("ovs_set_metadata: success: metat file %s is removed" % (cfg_file))
            except OSError, err:
                if err.errno != errno.ENOENT:
                    raise
            return "metat file %s is removed" % (cfg_file)

        # always over write exisiting data
        f = codecs.open(cfg_file, "w", encoding="utf-8")
        f.write("#This file was dynamically created by OVM manager. Please Do not edit\n")
        f.write("METADATA=%s\n" % (data))
        f.close()
        logging.info("ovs_set_metadata: success: set metatdata %s for interface %s" % (data, interface))
        return "set meta data %s for interface %s" % (data, interface)

    def ovs_change_mtu(self, interface, mtu):
        """
        Changes MTU on a physical,vlan,bond,and bridge interface.
        Changing a bond MTU will also change the MTU of its slaves.
        Changing the MTU of an interface that is part of a bridge,
        will cause the bridge MTU and all of the interfaces change.
        When a Guest VIF attach to a bridge,the VIF MTU will be set
        to the bridge MTU


        @param   interface    Physical,bond,vlan, and a bridge
        @param   MTU          Values are 1500 to 64000

        @return  If successful, returns the interface, and the new MTU
                 Raises an exception on failure

        Restriction:
        -Can not change the MTU of a bridge without interfaces.
        -VLAN MTU must less or equal to the MTU of the underlying
         physical interface.
        """

        if mtu < "1500" and mtu > "64000":
            raise Exception ("MTU %s is out of range. MTU values must be (1500 to 64000)" % mtu)
        cmd = ["%s/ovs-change-mtu" % self.script_dir, "netdev=%s" % interface, "mtu=%d" % int(mtu)]
        run_cmd(cmd)
        logging.info("ovs-change-mtu: success. interface %s set to %s" % (interface, mtu))
        return "interface %s set to %s" % (interface, mtu)

    def ovs_ovca_network(self, bond, *args):
        """
        Re-configure the management interface network on OVCA compute nodes.It works as follow:
        - Removes current bond interface(s)
        - Adds the PVI interfaces to the bond
        - Flushes current bond addreses (DHCP,Static)
        - Adds the new given address, and default gateway if present, otherwise use the current one.

        NOTE: This command should be used with ovs_async_proc

        @param bond   One of the logical channel bonds (bond0, bond1 ...etc)
        @param args   Is a list of the following arguments:

              primary    Primary physical or PVI interface (eth0,eth1 ..etc).
              backup     Backup physical or PVI interface  (eth0,eth1 ..etc)
                         Primary/backup are only valid for active-backup mode,
                         they are used as a place holder when configuring the other two modes.
              IP address IPv4 address in decimal notation (101.230.112)
              netmask    Standard netmask in a decimal notation,NOT CIDR.
                              example(255.255.255.0)
              gateway    Standard address in decimal notation (101.230.1)
                         This is an optional parameter, if not present the current default 
                          gateway will be used.


        @return       If successful, returns the names of the physical interfaces and the bond
                      Raises an exception on failure
        Restriction:
        -bond     must be one of the logical channel bonds (bond0, bond1 ...etc)
        -primary  must be physical/PVI, and not be member of a bridge, or slave to a bond
        -backup   must be physical/PVI, and not be member of a bridge, or slave to a bond

        """
        if len(args) < 4:
            raise Exception ("ovs_ovca_network failed: primary, backup, IP address, and netmask are required")
        primary = args[0]
        backup  = args[1]
        ipaddr  = args[2]
        netmask = args[3]
 
        #Validate IP
        try:
            socket.inet_aton(ipaddr)
            seprator = ipaddr.split(".")
        except:
            raise Exception ("IP %s is not valid." % ipaddr)
        if len(seprator) != 4:
                raise Exception ("ovs_ovca_network faild: IP %s is not valid." % ipaddr)

        #Validate netmask
        seprator = netmask.split(".")
        if len(seprator) != 4:
            raise Exception ("ovs_ovca_network faild: netmask %s is not legal valid." % netmask )


        if len(args) == 5:
            gateway = args[4]
            try:
                socket.inet_aton(gateway)
                seprator = ipaddr.split(".")
            except:
                raise Exception ("IP %s is not valid." % ipaddr)
            if len(seprator) != 4:
                raise Exception ("ovs_ovca_network faild: gateway %s is not valid." % gateway)
        else:
            gateway ="" 

        cmd = ["%s/ovs-ovca-network" % self.script_dir, "bondm=%s" % bond, "primary=%s" % primary, 
                     "backup=%s" % backup, "ip=%s" % ipaddr, "mask=%s" % netmask, "gateway=%s" % gateway]
        run_cmd(cmd)
        logging.info("ovs_ovca_network: success. bonding interfaces %s %s to %s" % (primary, backup, bond))

    def discover_network(self):
        """
        Discover information about the current network configuration.
        This includes the state of physical NICs, bonds, and bridges.  Also return
        information stored for this server that is needed to configure
        the network when the OVM Manager is not available.

        Discovery of the current network configuration is handled by invoking
        a python extension that calls legacy C code from the VI agent.

        @param None

        @return    Returns the discovery data as an XML document.
                   Raises an exception on failure.
        """
        lock = filelock.FileLock('/var/run/ovs-agent/discover_network.lock')
        lock.acquire(wait=120)
        try:
            (status, network_xml, logmsg) = OVSDiscoverExtension.discover_network(0)
            if status != 0:
                raise Exception(logmsg)
            result = '<Discover_Network_Result><Network><Active>%s</Active></Network></Discover_Network_Result>' % network_xml
            doc = xml.dom.minidom.parseString(result)
            for node in doc.getElementsByTagName("Device"):
                device = node.getAttribute("Name")
                config = linux_netutils.getConfig(device)
                if "BOOTPROTO" in config and config["BOOTPROTO"] in ["dhcp", "none", "static"]:
                    proto = doc.createElement("BootProto")
                    proto_text = doc.createTextNode(config["BOOTPROTO"])
                    proto.appendChild(proto_text)
                    node.appendChild(proto)
                meta = self.get_meta(device)
                if meta:
                    metadata = doc.createElement("MetaData")
                    metadata_text = doc.createTextNode(meta)
                    metadata.appendChild(metadata_text)
                    node.appendChild(metadata)
        finally:
            lock.release()
        return doc.toxml()

    def get_meta(self, device):
        filename = '/etc/sysconfig/network-scripts/meta-%s' % device
        if os.path.exists(filename):
            metafile = codecs.open(filename, encoding='utf-8')
            meta = metafile.readlines()
            metafile.close()
            for line in meta:
                if line.startswith('METADATA='):
                    return line[len('METADATA='):]
