# Copyright (C) 2006-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import collections
import Cookie
import httplib
import json
import logging
import os
import pickle
import signal
import struct
import threading
import time
import urlparse
from SocketServer import ThreadingUnixStreamServer, StreamRequestHandler
from agent.lib.constants import (LOCAL_SERVER_DB, NOTIFICATION_SERVER_ADDRESS,
        SSL_SERVER_PRIVATE_KEY, SSL_SIGNED_SERVER_CERTIFICATE)
from agent.lib.daemon import ignore_eintr, init_log
from agent.lib.db import read_item


MANGER_UUID = None
EVENT_QUEUE = collections.deque(maxlen=1024)
STATISTIC_QUEUE = collections.deque(maxlen=1024)
EVENT_SESSION_RESET = True
STATISTIC_SESSION_RESET = True


class Session(httplib.HTTPSConnection):

    def __init__(self, url):
        url_info = urlparse.urlparse(url)
        httplib.HTTPSConnection.__init__(
            self,
            host=url_info.hostname,
            port=url_info.port,
            key_file=SSL_SERVER_PRIVATE_KEY,
            cert_file=SSL_SIGNED_SERVER_CERTIFICATE)
        self.url = url_info.path
        self.headers = {'Accept': 'application/json',
                        'Content-Type': 'application/json'}
        self.cookies = {}

    def post(self, data=None):
        try:
            #
            # Send the post request. If the connection has been dropped
            # then we will get an exception. The exception can occur
            # when we send the request (CannotSendRequest) or when we
            # read the reply (BadStatusLine) depending when data are
            # effectively flushed to the socket.
            #
            return self._do_post(data)
        except httplib.HTTPException:
            #
            # The request has failed. We just close the connection and
            # the next request will automatically re-open it.
            #
            self.close()

        return self._do_post(data)

    def _do_post(self, data):
        self.request("POST", self.url,
                     json.dumps(data, sort_keys=True),
                     headers=self.headers)
        response = self.getresponse()
        content = response.read()

        if response.status != httplib.OK:
            raise Exception('%s %s' % (response.reason, content))

        #
        # We manage cookies very basically: any cookie returned by the
        # request is accumulated into self.cookies. Then we use all
        # cookies present in self.cookies in the next request. We don't
        # take any cookie attribute (Expires, Path...) into account.
        #
        cookies = Cookie.SimpleCookie()
        for hname, hvalue in response.getheaders():
            if hname.lower() == "set-cookie":
                cookies.load(hvalue)

        for cname, cvalue in cookies.items():
            self.cookies[cname] = cvalue

        self.headers['Cookie'] = "; ".join(v.OutputString(attrs=[])
                                           for v in self.cookies.values())
        if not self.headers['Cookie']:
            del self.headers['Cookie']

        return content


class MsgHandler(StreamRequestHandler):
    def handle(self):
        while True:
            try:
                chunk = self.connection.recv(4)
                if len(chunk) < 4:
                    break
                slen = struct.unpack(">L", chunk)[0]
                chunk = self.connection.recv(slen)
                while len(chunk) < slen:
                    chunk = chunk + self.connection.recv(slen - len(chunk))
                message_type, message, data = pickle.loads(chunk)
                if message_type == 'event':
                    if len(EVENT_QUEUE) >= 1024:
                        logging.warn('Event queue is full. Drop new event: %s %s', message, data)
                    else:
                        EVENT_QUEUE.appendleft((message, data))
                if message_type == 'statistic':
                    STATISTIC_QUEUE.appendleft(message)
            except Exception, err:
                logging.error('Error reading message: %s', err)


class MsgServer(ThreadingUnixStreamServer):

    allow_reuse_address = True

    def __init__(self, addr):
        if os.path.exists(addr):
            os.unlink(addr)
        ThreadingUnixStreamServer.__init__(self, addr, MsgHandler)

    def serve_forever(self, poll_interval=0.5):
        while True:
            try:
                ignore_eintr(MsgServer.handle_request, self)
            except Exception, err:
                logging.debug('Error handle request: %s', err)


def sleep(interval):
    while interval > 1:
        if EVENT_SESSION_RESET or STATISTIC_SESSION_RESET:
            return
        time.sleep(1)
        interval -= 1
    time.sleep(interval)


def queue_remove(queue, items):
    for item in items:
        try:
            if item == queue[-1]:
                queue.pop()
        except Exception, err:
            logging.error('Error remove %s: %s', item, err)
            break


def event_sender():
    global EVENT_SESSION_RESET
    session = None
    while True:
        try:
            if EVENT_SESSION_RESET:
                try:
                    session.close()
                except Exception:
                    pass
                session = None
                EVENT_SESSION_RESET = False
            if session is None:
                url = read_item(LOCAL_SERVER_DB, 'manager_event_url')
                if not url:
                    sleep(30)
                    continue
                if not os.path.exists(SSL_SIGNED_SERVER_CERTIFICATE):
                    sleep(30)
                    continue
                session = Session(url)
            if not EVENT_QUEUE:
                sleep(1)
                continue
            events = list(EVENT_QUEUE)
            events.reverse()
            events_formatted = [{'message': message, 'data': data}
                                for message, data in events]
            logging.debug('sent events: %s', events)
            session.post(events_formatted)
            queue_remove(EVENT_QUEUE, events)
        except Exception, err:
            try:
                session.close()
            except Exception:
                pass
            session = None
            logging.exception('Error sending events: %s', err)
            sleep(30)


def statistic_sender():
    global STATISTIC_SESSION_RESET
    session = None
    while True:
        try:
            if STATISTIC_SESSION_RESET:
                try:
                    session.close()
                except Exception:
                    pass
                session = None
                STATISTIC_SESSION_RESET = False
            if session is None:
                url = read_item(LOCAL_SERVER_DB, 'manager_statistic_url')
                if not url:
                    sleep(30)
                    continue
                if not os.path.exists(SSL_SIGNED_SERVER_CERTIFICATE):
                    sleep(30)
                    continue
                session = Session(url)
            if not STATISTIC_QUEUE:
                sleep(1)
                continue
            statistics = list(STATISTIC_QUEUE)
            statistics.reverse()
            #logging.debug('sent statistics: %s', statistics)
            session.post(statistics)
            queue_remove(STATISTIC_QUEUE, statistics)
        except Exception, err:
            try:
                session.close()
            except Exception:
                pass
            session = None
            logging.error('Error sendinging statistics: %s', err)
            sleep(30)


def sig_handler(sig, frame):
    if sig == signal.SIGHUP:
        manager_uuid = read_item(LOCAL_SERVER_DB, 'manager_uuid')
        if manager_uuid != MANGER_UUID:
            EVENT_QUEUE.clear()
            STATISTIC_QUEUE.clear()
        global EVENT_SESSION_RESET
        global STATISTIC_SESSION_RESET
        EVENT_SESSION_RESET = True
        STATISTIC_SESSION_RESET = True


def serve_forever():
    logging.info('Notification Server started.')
    global MANGER_UUID
    MANGER_UUID = read_item(LOCAL_SERVER_DB, 'manager_uuid')
    signal.signal(signal.SIGHUP, sig_handler)
    threading.Thread(target=event_sender).start()
    threading.Thread(target=statistic_sender).start()
    unix_server = MsgServer(NOTIFICATION_SERVER_ADDRESS)
    unix_server.serve_forever()


if __name__ == '__main__':
    init_log()
    serve_forever()
