# Copyright (C) 2006-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import glob
import logging
import os
import re
import string
import socket
import signal

from agent.extension import OVSOCFS2Extension
from agent.lib import filelock
from agent.lib.db import dump_db
from agent.lib.constants import (CLUSTER_STATE_DLM_READY,
				 CLUSTER_STATE_HEARTBEATING,
				 CLUSTER_STATE_OFFLINE,
				 CLUSTER_STATE_ONLINE,
				 CLUSTER_STATE_POOLFS_MOUNTED,
				 CLUSTER_LOCK, POOLFS_MOUNT, LOCAL_SERVER_DB,
				 MEMBERSHIP_STATE_CLUSTERED,
				 NFS_MOUNT)
from agent.lib.settings import get_poolfs_uuid, get_membership_state
from agent.target.dlm import dlm_ready, dlm_leave_domain
from agent.target.host import run_cmd, retry_cmd, _get_ip_dev
from agent.lib.files import file_read

#
# Constants
#

# pool fs
POOLFS_IMG = "ovspoolfs.img"
POOLFS_IMG_SIZE = 10240
POOLFS_LABEL = "OVS_POOL_FILESYSTEM"
POOLFS_NODE_SLOTS = "32"

# cluster
CLUSTER_CONFIGFS = "/sys/kernel/config/cluster"
CLUSTER_CONF = "/etc/ocfs2/cluster.conf"
CLUSTER_CONF_ORIG = "/etc/ocfs2/cluster.conf.orig"
CLUSTER_CONF_DIR = "/etc/ocfs2"
O2CB_CONF = "/etc/sysconfig/o2cb"


#
# Functions used by the agent generic framework.
#

def get_cluster_state():
    """
    Get the current state of the cluster.

    @param      None

    @return     One of "Offline", "Online", "Heartbeating",
                "PoolFS_Mounted" and "DLM_Ready".
    """
    state = CLUSTER_STATE_OFFLINE
    cluster_name = get_ocfs2_cluster_name()

    if cluster_name:
        state = CLUSTER_STATE_ONLINE
        poolfs_uuid = get_poolfs_uuid()

        if poolfs_uuid:
            mount_point = os.path.join(POOLFS_MOUNT, poolfs_uuid)
            if os.path.ismount(mount_point):
                if dlm_ready():
                    return CLUSTER_STATE_DLM_READY
                else:
                    return CLUSTER_STATE_POOLFS_MOUNTED

        output = run_cmd(["service", "o2cb", "status"])
        for line in output.splitlines():
            if line.strip() == "Checking O2CB heartbeat: Active":
                return CLUSTER_STATE_HEARTBEATING

    return state


#
# Functions used only by agent/cluster/o2cb.py and agent/cluster/ocfs2.py.
#

def get_ocfs2_cluster_name():
    """
    Get the ocfs2 cluster name in configfs.

    @return     Cluster name. Null string if not found.
    """
    ptn = os.path.join(CLUSTER_CONFIGFS, "*")
    lst = [ os.path.basename(e) for e in glob.glob(ptn) ]
    num = len(lst)

    if num == 0:
        return ""
    elif num == 1:
        return lst[0]
    else:
        raise Exception("Multiple clusters found: %s" % lst)


def get_cluster_ip():
    cn = get_ocfs2_cluster_name()
    if not cn:
        return ''
    hn = socket.gethostname()
    path = os.path.join(CLUSTER_CONFIGFS, cn, 'node', hn, 'ipv4_address')
    return file_read(path).strip()


def get_bridge_interface(bridge):
    cmd = ['brctl', 'show']
    output = run_cmd(cmd)
    for ln in output.splitlines()[1:]:
        fields = ln.split()
        if len(fields) > 0 and fields[0] == bridge:
            return fields[-1]
    return bridge


def get_cluster_interface():
    ip = get_cluster_ip()
    if not ip:
        return ''
    br = _get_ip_dev(ip)
    return get_bridge_interface(br)


def is_cluster_online():
    cmd = ["o2cb", "cluster-status"]
    try:
        # returns 0 if cluster online, 1 otherwise
        run_cmd(cmd)
        return True
    except Exception:
        return False


def server_cluster_debug():
    d = {}
    debug_paths = [
            "/sys/kernel/debug/o2dlm",
            "/sys/kernel/debug/o2hb",
            "/sys/kernel/debug/o2net",
            "/sys/kernel/debug/ocfs2",]
    for p in debug_paths:
        if os.path.exists(p):
            d[p] = os.listdir(p)
    cmd = ["service", "o2cb", "status"]
    d[" ".join(cmd)] = run_cmd(cmd)
    logging.debug("cluster debug: %s", d)


def server_cluster_online():
    """
    Register nodes and start heartbeating.

    @return     None. Raises an exception on failure.
    """
    server_cluster_debug()

    lock = filelock.FileLock(CLUSTER_LOCK)
    lock.acquire(wait=60)
    try:
        cmd = ["service", "o2cb", "online"]
        run_cmd(cmd)
    finally:
        lock.release()

    server_cluster_debug()

    if not is_cluster_online():
        raise Exception("Unable to online cluster.")


def server_cluster_offline():
    """
    Stop heatbeating and unregister nodes.

    @return     None. Raises an exception on failure.
    """
    server_cluster_debug()

    lock = filelock.FileLock(CLUSTER_LOCK)
    lock.acquire(wait=60)
    try:
        dlm_leave_domain()
        cmd = ["service", "o2cb", "offline"]
        run_cmd(cmd)
    finally:
        lock.release()

    server_cluster_debug()

    if is_cluster_online():
        raise Exception("Unable to offline cluster.")


def find_associated_loop_devices(filepath):
    associated_loop_devices = []
    try:
        output = run_cmd(['losetup', '-j', filepath])
        for line in output.splitlines():
            match = re.match(r'(/dev/loop\d+):.*', line)
            if match:
                associated_loop_devices.append(match.group(1))
    except Exception, err:
        logging.warn('Error finding associated loop devices: %s', err)
    return associated_loop_devices


def setup_loop_device(poolfs_img):
    associated_loop_devices = find_associated_loop_devices(poolfs_img)
    if associated_loop_devices:
        return associated_loop_devices[0]
    else:
        return run_cmd(['losetup', '--show', '-f', poolfs_img])


def detach_loop_device(poolfs_img):
    associated_loop_devices = find_associated_loop_devices(poolfs_img)
    if associated_loop_devices:
        run_cmd(['losetup', '-d', ' '.join(associated_loop_devices)])


def mount_device(device, mount_point, interval = 5, count = 3):
    logging.debug("Trying to mount %s to %s", device, mount_point)

    if not os.path.exists(mount_point):
        os.makedirs(mount_point)

    device = device.strip().rstrip("/")
    mount_point = mount_point.strip().rstrip("/")
    mounted = False

    for ln in open("/proc/mounts", "r"):
        dev, mnt = map(string.rstrip, ln.split()[0:2], ["/", "/"])
        if device == dev or mount_point == mnt:
            if device == dev and mount_point == mnt:
                mounted = True
                break
            else:
                raise Exception("%s already mounted on %s" % (dev, mnt))

    if mounted:
        cmd = [ "mount", "-o", "remount", device ]
        retry_cmd(cmd, interval, count)
        logging.debug("%s re-mounted to %s", device, mount_point)
    else:
        cmd = [ "mount", device, mount_point ]
        retry_cmd(cmd, interval, count)
        logging.debug("%s mounted to %s", device, mount_point)


def umount_device(device):
    logging.debug("Trying to umount %s", device)

    device = device.strip().rstrip("/")
    mount_point = None

    for ln in open("/proc/mounts", "r"):
        dev, mnt = map(string.rstrip, ln.split()[0:2], ["/", "/"])
        if device == dev or device == mnt:
            mount_point = mnt
            break

    if not mount_point:
        logging.debug("%s not mounted, skip umounting", device)
        return

    # force unmount, in case of an unreachable NFS system
    try:
        retry_cmd(["umount", "-f", mount_point])
    except Exception:
        # list all processes accessing files on the filesystem
        try:
            output = run_cmd(["lsof", mount_point]).strip()
            logging.error("Device or resource busy: \n%s", output)
        except Exception:
            pass
        raise

    logging.debug("%s umounted", device)

    # remove mount point if empty
    try:
        os.rmdir(mount_point)
    except Exception, err:
        logging.warn("Error removing mount point: %s", err)


def init_pool_filesystem():
    if (get_membership_state() != MEMBERSHIP_STATE_CLUSTERED or
        get_cluster_state() == CLUSTER_STATE_DLM_READY):
        return

    def timeout_handler(signum, frame):
        logging.debug("init_pool_filesystem error: timeout occurred")
        raise RuntimeError('Timeout')

    d = dump_db(LOCAL_SERVER_DB)
    poolfs_type = d["poolfs_type"]
    poolfs_nfsbase_uuid = d["poolfs_nfsbase_uuid"]
    poolfs_target = d["poolfs_target"]
    poolfs_uuid = d["poolfs_uuid"]
    poolfs_mount_point = os.path.join(POOLFS_MOUNT, poolfs_uuid)

    if poolfs_type == "nfs":
        try:
            wait_time = 45
            logging.debug("init_pool_filesystem(nfs): starting timer - max %d seconds", wait_time)
            signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(wait_time)

            nfsbase_mount_point = os.path.join(NFS_MOUNT, poolfs_nfsbase_uuid)
            poolfs_img = os.path.join(nfsbase_mount_point, POOLFS_IMG)
            mount_device(poolfs_target, nfsbase_mount_point, 5, 1)

            poolfs_device = setup_loop_device(poolfs_img)

            server_cluster_online()
            mount_device(poolfs_device, poolfs_mount_point, 5, 1)

            signal.alarm(0)

            logging.debug("init_pool_filesystem(nfs): stopped timer")

        except Exception, err:
            logging.debug("init_pool_filesystem: mount nfs pool FS failed - error %s", err)
  
            try:
                server_cluster_offline()
            except Exception, e:
                logging.debug("init_pool_filesystem: server_cluster_offline failed - error %s", e)

            try:
                detach_loop_device(poolfs_img)
            except Exception, e:
                logging.debug("init_pool_filesystem: detach_loop_device failed - error %s", e)
            try:
                umount_device(poolfs_target)
            except Exception, e:
                logging.debug("init_pool_filesystem: umount_device failed - error %s", e)

            raise err
            
    else:
        try:
            wait_time = 30
            logging.debug("init_pool_filesystem(phy): starting timer - max %d seconds", wait_time)
            signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(wait_time)

            server_cluster_online()
            mount_device(poolfs_target, poolfs_mount_point, 5, 1)
  
            signal.alarm(0)
            logging.debug("init_pool_filesystem(phy): stopped timer")

        except Exception, err:
            logging.debug("init_pool_filesystem: mount physical pool FS failed - error %s", err)

            try:
                server_cluster_offline()
            except Exception, e:
                logging.debug("init_pool_filesystem: server_cluster_offline failed - error %s", e)

            raise err


#
# Functions using the OVSOCFS2Extension.
#

def ocfs2_file_lookup(device, filename):
    """
    @return         Zero on success.
                    OCFS2_ET_FILE_NOT_FOUND (-1512559333L) on failure.
    """
    fs = OVSOCFS2Extension.open(device)
    try:
        return OVSOCFS2Extension.lookup(fs, filename)
    finally:
        OVSOCFS2Extension.close(fs)

def ocfs2_file_read(device, filename):
    """
    @return     File contents as a string.
                Raises an exception on failure.
    """
    fs = OVSOCFS2Extension.open(device)
    try:
        return OVSOCFS2Extension.file_read(fs, filename)
    finally:
        OVSOCFS2Extension.close(fs)

def ocfs2_file_write(device, filename, buf):
    """
    @return     Zero. Raises an exception on failure.
    """
    fs = OVSOCFS2Extension.open(device)
    try:
        return OVSOCFS2Extension.file_write(fs, filename, buf)
    finally:
        OVSOCFS2Extension.close(fs)

def ocfs2_file_remove(device, filename):
    """
    @return     Zero. Raises an exception on failure.
    """
    fs = OVSOCFS2Extension.open(device)
    try:
        return OVSOCFS2Extension.file_remove(fs, filename)
    finally:
        OVSOCFS2Extension.close(fs)
