#!/usr/bin/env python
#
# Copyright (C) 2011-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for dealing with vdisks
"""

import st_utils

import os
import traceback
import datetime

try:
	os.SEEK_SET	# defined in python 2.5
except AttributeError:
	os.SEEK_SET, os.SEEK_CUR, os.SEEK_END	= range( 3 )

####################################################################################################

__all__ = [	'VHDfileInfo_OBJ'
		,	'get_vhd_starting_datetime_UTC'
		,	'convert_vhd_timestamp_to_UTC'
		,	'calculate_VHD_checksum'
		,	'get_vhd_ftr_layout'
		,	'get_vhd_dyn_hdr_layout'
		,	'extract_vhd_info'
		,	'print_vhd_internals' ]

####################################################################################################

class VHDfileInfo_OBJ:

	def __init__( self, filePath, ftrLayout ):

		self.mFilePath							= filePath
		self.mFileSizeInBytes					= None
		self.mMaxDataSizeInBytes				= None

		self.mFtrOffset							= None
		self.mFtrBuffer							= None
		self.mFtrLayout							= ftrLayout
		self.mFtrCalculatedCkSum				= None
		self.mFtrCksumIsValid					= None
		self.mFtrDiskType						= None

		self.mDynHdrOffset						= None
		self.mDynHdrBuffer						= None
		self.mDynHdrLayout						= None
		self.mDynHdrCalculatedCkSum				= None
		self.mDynHdrCksumIsValid				= None
		self.mDynBATsizeInBytes					= None
		self.mDynHdrDataBlockBitmapSizeInBits	= None
		self.mDynHdrDataBlockBitmapSizeInBytes	= None
		self.mDynHdrTotalBlockSizeInBytes		= None
		self.mDynHdrTotalBlockSizeInSectors		= None

		self.mFtrCopyOffset						= None
		self.mFtrCopyIsConsistent				= None	# == mFtr when Dynamic sparse VHD file

	def __repr__( self ):

		return "fP: %s fS: %s dT: %s" % \
					(	self.mFilePath
					,	self.mFileSizeInBytes
					,	self.mFtrDiskType )

####################################################################################################
def get_vhd_starting_datetime_UTC():
	"""
	This routine
	"""
#

							# January 1, 2000 12:00:00AM in UTC/GMT
	return	datetime.datetime(	2000,     1,   1,    0,      0,      0,           0, st_utils.VI_UTC() )
							# ( year, month, day, hour, minute, second, microsecond, tzinfo )

####################################################################################################
def convert_vhd_timestamp_to_UTC( vi_errorAndLog_obj, vhd_time_stamp ):
	"""
	This routine
	"""
#
	seconds_per_day					=  60 * 60 * 24
	vhd_time_stamp_day_truncated	= vhd_time_stamp / seconds_per_day
	vhd_time_stamp_sec_remainder	= vhd_time_stamp - ( vhd_time_stamp_day_truncated * seconds_per_day )
	vhd_time_stamp_timedelta		= datetime.timedelta( vhd_time_stamp_day_truncated, vhd_time_stamp_sec_remainder )

	vhd_starting_datetime_UTC		= get_vhd_starting_datetime_UTC()

	vhd_time_stamp_as_datetime_UTC	= vhd_starting_datetime_UTC + vhd_time_stamp_timedelta

	return vhd_time_stamp_as_datetime_UTC

####################################################################################################
def _INTERNAL__modify_checksum( calculated_checksum, vhd_byte, isAddVsSub ):
	"""
	This routine
	"""
#
	vhd_byte_asInt = ord( vhd_byte )

	if isAddVsSub:
		calculated_checksum	+= vhd_byte_asInt
	else:
		calculated_checksum	-= vhd_byte_asInt

	# Handle unsigned 32-bit wrap
	if calculated_checksum < 0:
		calculated_checksum	+= 0xFFFFFFFFL
	if calculated_checksum > 0xFFFFFFFFL:
		calculated_checksum	-= 0xFFFFFFFFL

	return calculated_checksum

####################################################################################################
def calculate_VHD_checksum( vi_errorAndLog_obj, vhd_byte_buffer, embedded_checksum_offset ):
	"""
	This routine
	"""
#
	calculated_checksum	= 0

	for vhd_byte in vhd_byte_buffer:
		calculated_checksum	= _INTERNAL__modify_checksum( calculated_checksum, vhd_byte, True )

	calculated_checksum	= _INTERNAL__modify_checksum( calculated_checksum, vhd_byte_buffer[ embedded_checksum_offset     ], False )
	calculated_checksum	= _INTERNAL__modify_checksum( calculated_checksum, vhd_byte_buffer[ embedded_checksum_offset + 1 ], False )
	calculated_checksum	= _INTERNAL__modify_checksum( calculated_checksum, vhd_byte_buffer[ embedded_checksum_offset + 2 ], False )
	calculated_checksum	= _INTERNAL__modify_checksum( calculated_checksum, vhd_byte_buffer[ embedded_checksum_offset + 3 ], False )

	calculated_checksum	= ~calculated_checksum & 0xFFFFFFFFL

	return calculated_checksum

####################################################################################################
def get_vhd_ftr_layout( vi_errorAndLog_obj ):
	"""
	This routine builds the vhd_ftr_layout
	"""
#
	vhd_ftr_frmtFieldname_TPL_LIST = [
			(	st_utils.vi_strctFrmt__bytes( 8 ),		'cookie'					)	# offset ==   0 |   8 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'features'					)	# offset ==   8 |   4 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'file_format_version'		)	# offset ==  12 |   4 bytes
		,	(	st_utils.vi_strctFrmt__be__u64(),		'data_offset'				)	# offset ==  16 |   8 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'time_stamp'				)	# offset ==  24 |   4 bytes
		,	(	st_utils.vi_strctFrmt__bytes( 4 ),		'creator_application'		)	# offset ==  28 |   4 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'creator_version'			)	# offset ==  32 |   4 bytes
		,	(	st_utils.vi_strctFrmt__bytes( 4 ),		'creator_host_os'			)	# offset ==  36 |   4 bytes
		,	(	st_utils.vi_strctFrmt__be__u64(),		'original_size'				)	# offset ==  40 |   8 bytes
		,	(	st_utils.vi_strctFrmt__be__u64(),		'current_size'				)	# offset ==  48 |   8 bytes
		,	(	st_utils.vi_strctFrmt__bytes( 4 ),		'disk_geometry'				)	# offset ==  56 |   4 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'disk_type'					)	# offset ==  60 |   4 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'checksum'					)	# offset ==  64 |   4 bytes
		,	(	st_utils.vi_strctFrmt__bytes( 16 ),		'unique_ID'					)	# offset ==  68 |  16 bytes
		,	(	st_utils.vi_strctFrmt__bytes( 1 ),		'saved_state'				)	# offset ==  84 |   1 bytes
		,	(	st_utils.vi_strctFrmt__bytes( 427 ),	'reserved'					)	# offset ==  85 | 427 bytes
		]																				# Total ==> 512 bytes ( fixed size hdr )

	vhd_ftr_layout = st_utils.vi_structLayout( vhd_ftr_frmtFieldname_TPL_LIST, 512 )

	return vhd_ftr_layout

####################################################################################################
def get_vhd_dyn_hdr_layout( vi_errorAndLog_obj ):
	"""
	This routine builds the vhd_dyn_hdr_layout
	"""
#
	vhd_dyn_hdr_frmtFieldname_TPL_LIST = [
			(	st_utils.vi_strctFrmt__bytes( 8 ),		'cookie'					)	# offset ==   0 |   8 bytes
		,	(	st_utils.vi_strctFrmt__be__u64(),		'data_offset'				)	# offset ==   8 |   8 bytes
		,	(	st_utils.vi_strctFrmt__be__u64(),		'table_offset'				)	# offset ==  12 |   8 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'header_version'			)	# offset ==  16 |   4 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'max_table_enteries'		)	# offset ==  24 |   4 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'block_size'				)	# offset ==  32 |   4 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'checksum'					)	# offset ==  36 |   4 bytes
		,	(	st_utils.vi_strctFrmt__bytes( 16 ),		'parent_unique_ID'			)	# offset ==  68 |  16 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'parent_time_stamp'			)	# offset ==  40 |   4 bytes
		,	(	st_utils.vi_strctFrmt__be__u32(),		'reserved'					)	# offset ==  40 |   4 bytes
		,	(	st_utils.vi_strctFrmt__bytes( 512 ),	'parent_unicode_name'		)	# offset ==  68 | 512 bytes
		,	(	st_utils.vi_strctFrmt__bytes( 192 ),	'parent_locator_entries'	)	# offset ==  68 | 192 bytes
		,	(	st_utils.vi_strctFrmt__bytes( 256 ),	'reserved2'					)	# offset ==  85 | 256 bytes
		]																				# Total ==> 1024 bytes ( fixed size hdr )

	vhd_dyn_hdr_layout = st_utils.vi_structLayout( vhd_dyn_hdr_frmtFieldname_TPL_LIST, 1024 )

	return vhd_dyn_hdr_layout

####################################################################################################
def extract_vhd_info( vi_errorAndLog_obj, filePath ):
	"""
	This routine

	Return:		VHDfileInfo_OBJ()
	"""
#
	if not os.path.exists( filePath ):
		vi_errorAndLog_obj.vi_print_err( "vhd file '%s' does not exist" % ( filePath ) )
		return None

	vhdFileInfo	= VHDfileInfo_OBJ( filePath[:], get_vhd_ftr_layout( vi_errorAndLog_obj ) )
	try:
		givenVHD_FILE					= open( vhdFileInfo.mFilePath, 'rb' )
		vhdFileInfo.mFileSizeInBytes	= os.path.getsize( vhdFileInfo.mFilePath )
	except:
#		vi_errorAndLog_obj.vi_croak_exception( traceback.format_exc(), "open( %s ) failed" % ( vhdFileInfo.mFilePath ) )
		vi_errorAndLog_obj.vi_print_err( "open( %s ) failed :\n%s" % ( vhdFileInfo.mFilePath, traceback.format_exc() ) )
		return None

	if vhdFileInfo.mFileSizeInBytes < 512:
		vi_errorAndLog_obj.vi_print_err( "Give file '%s' is too small %s" % ( vhdFileInfo.mFilePath, vhdFileInfo.mFileSizeInBytes ) )
		givenVHD_FILE.close()
		return None

	ftrSizeInBytes	= vhdFileInfo.mFileSizeInBytes % 512
	if ftrSizeInBytes == 0:
		ftrSizeInBytes	= 512
		zeroPadBytes	= 0
	else:
		if ftrSizeInBytes != 511:
			vi_errorAndLog_obj.vi_print_err( "Give file %s size %s is not an even 512 multiple" % ( vhdFileInfo.mFilePath, vhdFileInfo.mFileSizeInBytes ) )
			givenVHD_FILE.close()
			return None
		zeroPadBytes	= 512 - ftrSizeInBytes

	try:

		vhdFileInfo.mFtrOffset	= vhdFileInfo.mFileSizeInBytes - ftrSizeInBytes
		givenVHD_FILE.seek( vhdFileInfo.mFtrOffset, os.SEEK_SET )
		vhdFileInfo.mFtrBuffer	= givenVHD_FILE.read( ftrSizeInBytes )

		if zeroPadBytes:
			vhdFileInfo.mFtrBuffer	= vhdFileInfo.mFtrBuffer + chr(0) * zeroPadBytes

		vhdFileInfo.mFtrLayout.IngestBuffer( vhdFileInfo.mFtrBuffer )

		vhdFileInfo.mFtrDiskType	= vhdFileInfo.mFtrLayout.Get( 'disk_type' )
		vhd_ftr__checksum			= vhdFileInfo.mFtrLayout.Get( 'checksum' )

		vhdFileInfo.mFtrCalculatedCkSum	= calculate_VHD_checksum( vi_errorAndLog_obj, vhdFileInfo.mFtrBuffer, 64 )
		if vhd_ftr__checksum == vhdFileInfo.mFtrCalculatedCkSum:
			vhdFileInfo.mFtrCksumIsValid	= True
		else:
			vhdFileInfo.mFtrCksumIsValid	= False

		if vhdFileInfo.mFtrDiskType == 2:		# FxdVHD

			vhdFileInfo.mMaxDataSizeInBytes	= vhdFileInfo.mFtrLayout.Get( 'current_size' )

		elif vhdFileInfo.mFtrDiskType == 3:		# DynVHD

			vhdFileInfo.mFtrCopyOffset	= 0
			givenVHD_FILE.seek( vhdFileInfo.mFtrCopyOffset, os.SEEK_SET )
			copy_vhd_ftr_buf			= givenVHD_FILE.read( 512 )
			if copy_vhd_ftr_buf == vhdFileInfo.mFtrBuffer:
				vhdFileInfo.mFtrCopyIsConsistent	= True
			else:
				vhdFileInfo.mFtrCopyIsConsistent	= False
			del copy_vhd_ftr_buf

			vhdFileInfo.mDynHdrLayout	= get_vhd_dyn_hdr_layout( vi_errorAndLog_obj )

			vhdFileInfo.mDynHdrOffset				= 512
			givenVHD_FILE.seek( vhdFileInfo.mDynHdrOffset, os.SEEK_SET )
			vhdFileInfo.mDynHdrBuffer				= givenVHD_FILE.read( 1024 )

			vhdFileInfo.mDynHdrLayout.IngestBuffer( vhdFileInfo.mDynHdrBuffer )

			vhdFileInfo.mDynHdrDataBlockSizeInBytes	= vhdFileInfo.mDynHdrLayout.Get( 'block_size' )
			vhd_dyn_hdr__max_table_enteries			= vhdFileInfo.mDynHdrLayout.Get( 'max_table_enteries' )
			vhd_dyn_hdr__checksum					= vhdFileInfo.mDynHdrLayout.Get( 'checksum' )

			vhdFileInfo.mDynBATsizeInBytes	= vhd_dyn_hdr__max_table_enteries * 4
			vhdFileInfo.mDynBATsizeInBytes	= st_utils.vi_align_to_next( vhdFileInfo.mDynBATsizeInBytes, 512 )

			vhdFileInfo.mDynHdrCalculatedCkSum		= calculate_VHD_checksum( vi_errorAndLog_obj, vhdFileInfo.mDynHdrBuffer, 36 )
			if vhd_dyn_hdr__checksum == vhdFileInfo.mDynHdrCalculatedCkSum:
				vhdFileInfo.mDynHdrCksumIsValid	= True
			else:
				vhdFileInfo.mDynHdrCksumIsValid	= False

			vhdFileInfo.mDynHdrDataBlockBitmapSizeInBits		 = vhdFileInfo.mDynHdrDataBlockSizeInBytes / 512		# sectors to bits

			vhdFileInfo.mDynHdrDataBlockBitmapSizeInBytes		 = vhdFileInfo.mDynHdrDataBlockBitmapSizeInBits / 8		# bits to bytes
			if vhdFileInfo.mDynHdrDataBlockBitmapSizeInBits % 8:
				vhdFileInfo.mDynHdrDataBlockBitmapSizeInBytes	+= 1
			vhdFileInfo.mDynHdrDataBlockBitmapSizeInBytes		 = st_utils.vi_align_to_next( vhdFileInfo.mDynHdrDataBlockBitmapSizeInBytes, 512 )	# align up to sector size

			vhdFileInfo.mDynHdrTotalBlockSizeInBytes	= vhdFileInfo.mDynHdrDataBlockBitmapSizeInBytes + vhdFileInfo.mDynHdrDataBlockSizeInBytes
			vhdFileInfo.mDynHdrTotalBlockSizeInSectors	= vhdFileInfo.mDynHdrTotalBlockSizeInBytes / 512

			vhdFileInfo.mMaxDataSizeInBytes		= vhdFileInfo.mDynHdrDataBlockSizeInBytes * vhd_dyn_hdr__max_table_enteries

		else:

			vhdFileInfo.mMaxDataSizeInBytes	= None

	except:
#		vi_errorAndLog_obj.vi_croak_exception( traceback.format_exc(), "extract_vhd_info( %s ) extraction failed" % ( vhdFileInfo.mFilePath ) )
		vi_errorAndLog_obj.vi_print_err( "extract_vhd_info( %s ) extraction failed :\n%s" % ( vhdFileInfo.mFilePath, traceback.format_exc() ) )
		givenVHD_FILE.close()
		return None

	givenVHD_FILE.close()

	return vhdFileInfo

####################################################################################################
def print_vhd_internals( vi_errorAndLog_obj, printFcn, vhdFileInfo, dumpVHDbuffersAsHex ):
	"""
	This routine
	"""
#
	givenVHDsizeInGB	= float(vhdFileInfo.mFileSizeInBytes) / 1000.0 / 1000.0 / 1000.0

	vhd_ftr__cookie					= vhdFileInfo.mFtrLayout.Get( 'cookie' )
	vhd_ftr__features				= vhdFileInfo.mFtrLayout.Get( 'features' )
	vhd_ftr__file_format_version	= vhdFileInfo.mFtrLayout.Get( 'file_format_version' )
	vhd_ftr__time_stamp				= vhdFileInfo.mFtrLayout.Get( 'time_stamp' )
	vhd_ftr__creator_application	= vhdFileInfo.mFtrLayout.Get( 'creator_application' )
	vhd_ftr__creator_version		= vhdFileInfo.mFtrLayout.Get( 'creator_version' )
	vhd_ftr__creator_host_os		= vhdFileInfo.mFtrLayout.Get( 'creator_host_os' )
	vhd_ftr__original_size			= vhdFileInfo.mFtrLayout.Get( 'original_size' )
	vhd_ftr__current_size			= vhdFileInfo.mFtrLayout.Get( 'current_size' )
	vhd_ftr__disk_geometry			= vhdFileInfo.mFtrLayout.Get( 'disk_geometry' )
	vhd_ftr__disk_type				= vhdFileInfo.mFtrLayout.Get( 'disk_type' )
	vhd_ftr__checksum				= vhdFileInfo.mFtrLayout.Get( 'checksum' )
	vhd_ftr__unique_ID				= vhdFileInfo.mFtrLayout.Get( 'unique_ID' )
	vhd_ftr__saved_state			= vhdFileInfo.mFtrLayout.Get( 'saved_state' )

	if vhdFileInfo.mFtrCksumIsValid:
		vhd_ftr__checksum_equivalence_str	= '=='
	else:
		vhd_ftr__checksum_equivalence_str	= '!='

	vhd_ftr__time_stamp_datetime_UTC	= convert_vhd_timestamp_to_UTC( vi_errorAndLog_obj, vhd_ftr__time_stamp )

	vhd_ftr__creator_application_asInt	= st_utils.vi_byteStr4ToUInt32( vhd_ftr__creator_application, False )
	vhd_ftr__creator_host_os_asInt		= st_utils.vi_byteStr4ToUInt32( vhd_ftr__creator_host_os, False )
	vhd_ftr__original_size_inGB			= float(vhd_ftr__original_size) / 1000.0 / 1000.0 / 1000.0
	vhd_ftr__current_size_inGB			= float(vhd_ftr__current_size)  / 1000.0 / 1000.0 / 1000.0

	if vhd_ftr__disk_type == 3:

		vhd_dyn_hdr__cookie					= vhdFileInfo.mDynHdrLayout.Get( 'cookie' )
		vhd_dyn_hdr__table_offset			= vhdFileInfo.mDynHdrLayout.Get( 'table_offset' )
		vhd_dyn_hdr__header_version			= vhdFileInfo.mDynHdrLayout.Get( 'header_version' )
		vhd_dyn_hdr__max_table_enteries		= vhdFileInfo.mDynHdrLayout.Get( 'max_table_enteries' )
		vhd_dyn_hdr__block_size				= vhdFileInfo.mDynHdrLayout.Get( 'block_size' )
		vhd_dyn_hdr__checksum				= vhdFileInfo.mDynHdrLayout.Get( 'checksum' )

		if vhdFileInfo.mDynHdrCksumIsValid:
			vhd_dyn_hdr__checksum_equivalence_str	= '=='
		else:
			vhd_dyn_hdr__checksum_equivalence_str	= '!='

		vhd_dyn_hdr__data_size				= vhd_dyn_hdr__block_size * vhd_dyn_hdr__max_table_enteries
		vhd_dyn_hdr__data_size_inGB			= float(vhd_dyn_hdr__data_size) / 1000.0 / 1000.0 / 1000.0

		if vhdFileInfo.mFtrCopyIsConsistent:
			vhd_dyn__ftr_copy_ftr_equivalence_str	= '=='
		else:
			vhd_dyn__ftr_copy_ftr_equivalence_str	= '!='

	printFcn( '' )
	printFcn( "FileSize:                %s [ %0.2f GB ] %s %s" % \
				(	vhdFileInfo.mFileSizeInBytes
				,	givenVHDsizeInGB
				,	vhdFileInfo.mFileSizeInBytes / 512
				,	vhdFileInfo.mFileSizeInBytes % 512 ) )

	if dumpVHDbuffersAsHex:
		printFcn( '' )
		printFcn( 'VHD ftr as hex:\n    ', None )
		idx	= 0
		for vhd_byte in vhdFileInfo.mFtrBuffer:
			printFcn( "%02X" % ( ord(vhd_byte) ), None )
			idx	+= 1
			if idx % 16 == 0:
				printFcn( '\n    ', None )

	printFcn( '' )
	printFcn( 'VHD ftr:' )
	printFcn( "    offset:              %s %s" % ( vhdFileInfo.mFtrOffset, vhdFileInfo.mFtrOffset / 512 ) )
	printFcn( "    cookie:              %s" % ( vhd_ftr__cookie ) )
	printFcn( "    features:            0x%08X" % ( vhd_ftr__features ) )
	printFcn( "    file_format_version: 0x%08X" % ( vhd_ftr__file_format_version ) )
	printFcn( "    time_stamp:          %s UTC [ %s ]" % ( vhd_ftr__time_stamp_datetime_UTC.strftime( "%Y-%m-%d.%H:%M:%S" ), vhd_ftr__time_stamp ) )
	printFcn( "    creator_application: 0x%08X [ %s ]" % ( vhd_ftr__creator_application_asInt, vhd_ftr__creator_application ) )
	printFcn( "    creator_version:     0x%08X" % ( vhd_ftr__creator_version ) )
	printFcn( "    creator_host_os:     0x%08X [ %s ]" % ( vhd_ftr__creator_host_os_asInt, vhd_ftr__creator_host_os ) )
	printFcn( "    original_size:       %s [ %0.2f GB ]" % ( vhd_ftr__original_size, vhd_ftr__original_size_inGB ) )
	printFcn( "    current_size:        %s [ %0.2f GB ]" % ( vhd_ftr__current_size, vhd_ftr__current_size_inGB ) )
	printFcn( "    disk_geometry:       0x%s" % ( st_utils.vi_as_hex( vhd_ftr__disk_geometry ) ) )
	printFcn( "    disk_type:           %s" % ( vhd_ftr__disk_type ) )
	printFcn( "    checksum:            0x%08X [ 0x%08X ] %s" % ( vhd_ftr__checksum, vhdFileInfo.mFtrCalculatedCkSum, vhd_ftr__checksum_equivalence_str ) )
	printFcn( "    unique_ID:           0x%s" % ( st_utils.vi_as_hex( vhd_ftr__unique_ID ) ) )
	printFcn( "    saved_state:         0x%s" % ( st_utils.vi_as_hex( vhd_ftr__saved_state ) ) )
	printFcn( '' )

	if vhd_ftr__disk_type == 3:

		if dumpVHDbuffersAsHex:
			printFcn( 'VHD dyn hdr as hex:\n    ', None )
			idx	= 0
			for vhd_byte in vhdFileInfo.mDynHdrBuffer:
				printFcn( "%02X" % ( ord(vhd_byte) ), None )
				idx	+= 1
				if idx % 16 == 0:
					printFcn( '\n    ', None )
			printFcn( '' )

		printFcn( 'VHD dyn hdr:' )
		printFcn( "    cookie:              %s" % ( vhd_dyn_hdr__cookie ) )
		printFcn( "    table_offset:        %s [ min: 1536 ] %s" % ( vhd_dyn_hdr__table_offset, vhd_dyn_hdr__table_offset / 512 ) )
		printFcn( "    header_version:      0x%08X" % ( vhd_dyn_hdr__header_version ) )
		printFcn( "    max_table_enteries:  %s" % ( vhd_dyn_hdr__max_table_enteries ) )
		printFcn( "    block_size:          %s" % ( vhd_dyn_hdr__block_size ) )
		printFcn( "    total_block_size:    %s [ ( %s ) %s + %s ] %s" % \
														(	vhdFileInfo.mDynHdrTotalBlockSizeInBytes
														,	vhdFileInfo.mDynHdrDataBlockBitmapSizeInBits
														,	vhdFileInfo.mDynHdrDataBlockBitmapSizeInBytes
														,	vhd_dyn_hdr__block_size
														,	vhdFileInfo.mDynHdrTotalBlockSizeInSectors ) )
		printFcn( "    data_size:           %s [ %0.2f GB ]" % ( vhd_dyn_hdr__data_size, vhd_dyn_hdr__data_size_inGB ) )
		printFcn( "    checksum:            0x%08X [ 0x%08X ] %s" % ( vhd_dyn_hdr__checksum, vhdFileInfo.mDynHdrCalculatedCkSum, vhd_dyn_hdr__checksum_equivalence_str ) )
		printFcn( "    ftr_vs_copy_ftr:     %s" % ( vhd_dyn__ftr_copy_ftr_equivalence_str ) )
		printFcn( '' )

####################################################################################################
# END
