#!/usr/bin/env python
#
# Copyright (C) 2011-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for parsing the network_config_directives.xml file.

<NetworkCfgDirectives>

	<!-- ================================================================== -->
	<!-- Standard DHCP example:                                             -->

<!--	<CfgNICdhcp>00:11:22:33:44:55</CfgNICdhcp>							-->

	<!-- ================================================================== -->
	<!-- Standard static IP example:                                        -->

<!--	<CfgNICstatic>														-->
<!--		<Interface>eth3</Interface>										-->
<!--		<StaticIP>192.168.1.2</StaticIP>								-->
<!--		<StaticIPmask>255.255.255.0</StaticIPmask>						-->
<!--	</CfgNICstatic>														-->

	<!-- ================================================================== -->
	<!-- MTU (jumbo frames) example:                                        -->

<!--	<CfgNICmtu>00:15:17:1C:99:64|4200</CfgNICmtu>						-->

	<!-- ================================================================== -->
	<!-- AOE (ATA over ethernet) example:                                   -->

<!--	<CfgNICaoe>00:15:17:1C:99:64</CfgNICaoe>							-->

	<!-- ================================================================== -->
	<!-- Configuration example of bonding.. Three types are available:      -->
	<!-- Failover, LinkAggregation, LoadBalance                           -->
	<!--                                                                    -->
	<!-- NOTE: The CfgBond MUST be before configuring the IP of the NIC.    -->
	<!-- The bond assumes first MAC address. Use bond0 instead of ethX to   -->
	<!-- configure IP.                                                      -->

<!--	<CfgBond>															-->
<!--		<BondTag>bond0</BondTag>										-->
<!--		<BondType>Failover</BondType>									-->
<!--		<NumOfInterfacesInBond>2</NumOfInterfacesInBond>				-->
<!--		<Interface>00:15:17:67:38:D6</Interface>						-->
<!--		<Interface>00:15:17:67:38:D7</Interface>						-->
<!--	</CfgBond>															-->

	<!-- ================================================================== -->
	<!-- Route example:  (see route man page for arg details)               -->

<!--	<CfgRoute>add -net 192.56.76.0 netmask 255.255.255.0 dev eth2</CfgRoute>	-->

	<!-- ================================================================== -->
	<!-- Ifconfig example:  (see ifconfig man page for arg details)         -->

<!--	<CfgIfconfig>eth2 down</CfgIfconfig>								-->

	<!-- ================================================================== -->
	<!-- Ethtool example:  (see ethtool man page for arg details)           -->

<!--	<CfgEthtool>-s eth0 speed 1000</CfgEthtool>							-->

	<!-- ================================================================== -->
	<!-- CONFIGURATION ATTRIBUTE:  extraTime                                -->
	<!--     Some HW configurations need extraTime to succeed               -->
	<!--         extraTime DHCP example:                                    -->

<!--	<CfgNICdhcp extraTime="60">00:11:22:33:44:55</CfgNICdhcp>			-->

</NetworkCfgDirectives>

"""

import st_utils

import os
import re
import traceback

import xml.dom.minidom

####################################################################################################

__all__ = [	'VI_NETWORK_DIRECTIVE__range'
		,	'VI_NETWORK_DIRECTIVE__DHCP'
		,	'VI_NETWORK_DIRECTIVE__STATIC_IP'
		,	'VI_NETWORK_DIRECTIVE__MTU'
		,	'VI_NETWORK_DIRECTIVE__AOE'
		,	'VI_NETWORK_DIRECTIVE__BOND'
		,	'VI_NETWORK_DIRECTIVE__ROUTE'
		,	'VI_NETWORK_DIRECTIVE__IFCONFIG'
		,	'VI_NETWORK_DIRECTIVE__ETHTOOL'
		,	'vi_network_directive__base'
		,	'vi_network_directive__DHCP__OBJ'
		,	'vi_network_directive__static_IP__OBJ'
		,	'vi_network_directive__MTU__OBJ'
		,	'vi_network_directive__AOE__OBJ'
		,	'vi_network_directive__bond__OBJ'
		,	'vi_network_directive__route__OBJ'
		,	'vi_network_directive__ifconfig__OBJ'
		,	'vi_network_directive__ethtool__OBJ'
		,	'vi_get_default_network_directive_fcn_list'
		,	'vi_extract_network_directive_list' ]

####################################################################################################

VI_NETWORK_DIRECTIVE__range			= ( 1, 9 )	# ( inclusive, exclusive ]
VI_NETWORK_DIRECTIVE__DHCP			= 1
VI_NETWORK_DIRECTIVE__STATIC_IP		= 2
VI_NETWORK_DIRECTIVE__MTU			= 3
VI_NETWORK_DIRECTIVE__AOE			= 4
VI_NETWORK_DIRECTIVE__BOND			= 5
VI_NETWORK_DIRECTIVE__ROUTE			= 6
VI_NETWORK_DIRECTIVE__IFCONFIG		= 7
VI_NETWORK_DIRECTIVE__ETHTOOL		= 8

class vi_network_directive__base:

	def GetDirectiveType( self ):
		vi_raise_not_implemented()

	def __init__( self ):
		self.mAttr_extraTime	= 0
		self.mFcn				= None

####################################################################################################

####################################################################################################
def _INTERNAL__extract_expected_xml_text( node_obj, expectedTagName ):
	"""
	This routine returns the text value for a particular element node
	"""
#
	if node_obj.nodeType != xml.dom.Node.ELEMENT_NODE:
		raise ValueError, "Given node_obj is not an ELEMENT_NODE : %d" % ( node_obj.nodeType )

	if node_obj.tagName != expectedTagName:
		raise ValueError, "Given node_obj.tagName '%s' does NOT match expected '%s'" % ( node_obj.tagName, expectedTagName )

	textNode_LIST	= []
	for child in node_obj.childNodes:
		if child.nodeType == xml.dom.Node.TEXT_NODE:
			textNode_LIST.append( child )
			continue
		if child.nodeType == xml.dom.Node.COMMENT_NODE:
			continue
		raise ValueError, "Given node_obj '%s' contains non-text node : %d : '%s'" % ( node_obj.tagName, child.nodeType, child.toxml() )

	if len( textNode_LIST ) != 1:
		raise ValueError, "Given node_obj '%s' does not have expected single text node : %d" % ( node_obj.tagName, len( textNode_LIST ) )

	return str(textNode_LIST[0].nodeValue)	# nodeValue is Unicode so convert to regular str

####################################################################################################

class vi_network_directive__DHCP__OBJ(vi_network_directive__base):

	def GetDirectiveType( self ):
		return VI_NETWORK_DIRECTIVE__DHCP

	def __init__( self, node_obj ):

		vi_network_directive__base.__init__( self )

		self.mInterface	= _INTERNAL__extract_expected_xml_text( node_obj, 'CfgNICdhcp' ).strip()

	def __repr__( self ):
		return "CfgNICdhcp: mInterface: %s" % ( self.mInterface )

####################################################################################################

class vi_network_directive__static_IP__OBJ(vi_network_directive__base):

	def GetDirectiveType( self ):
		return VI_NETWORK_DIRECTIVE__STATIC_IP

	def __init__( self, node_obj ):

		vi_network_directive__base.__init__( self )

		if node_obj.tagName != 'CfgNICstatic':
			raise ValueError, "Given node_obj.tagName '%s' does NOT match expected 'CfgNICstatic'" % ( node_obj.tagName )

		elemNode_LIST	= []
		for childNode in node_obj.childNodes:
			if childNode.nodeType == xml.dom.Node.ELEMENT_NODE:
				elemNode_LIST.append( childNode )
				continue
			if childNode.nodeType == xml.dom.Node.COMMENT_NODE:
				continue	# Ignore COMMENT_NODEs
			if childNode.nodeType == xml.dom.Node.TEXT_NODE:
				continue	# Ignore TEXT_NODE(whitespace)s
			raise ValueError, "Given CfgNICstatic contains non-text node : %d : '%s'" % ( childNode.nodeType, childNode.toxml() )

		if len( elemNode_LIST ) != 3:
			raise ValueError, "Given CfgNICstatic contains incorrect number of text nodes : %d" % ( len( elemNode_LIST ) )

		self.mInterface		= _INTERNAL__extract_expected_xml_text( elemNode_LIST[0], 'Interface' ).strip()
		self.mIPv4_addr		= _INTERNAL__extract_expected_xml_text( elemNode_LIST[1], 'StaticIP' ).strip()
		self.mIPv4_netmask	= _INTERNAL__extract_expected_xml_text( elemNode_LIST[2], 'StaticIPmask' ).strip()

	def __repr__( self ):
		return "CfgNICstatic: mInterface: %s mIPv4_addr: %s mIPv4_netmask: %s" % (
				self.mInterface
			,	self.mIPv4_addr
			,	self.mIPv4_netmask )

####################################################################################################

class vi_network_directive__MTU__OBJ(vi_network_directive__base):

	def GetDirectiveType( self ):
		return VI_NETWORK_DIRECTIVE__MTU

	def __init__( self, node_obj ):

		vi_network_directive__base.__init__( self )

		interfacePmtusize				= _INTERNAL__extract_expected_xml_text( node_obj, 'CfgNICmtu' ).strip()
		interfacePmtusize_cmpldRegEx	= re.compile( '^(\S+)\|(\d+)$' )
		interfacePmtusize_match			= interfacePmtusize_cmpldRegEx.match( interfacePmtusize )
		if not interfacePmtusize_match:
			raise ValueError, "MTU: child text was bad format : '%s'" % ( childNodes_LIST[0].nodeValue )

		self.mInterface		= interfacePmtusize_match.group( 1 )
		self.mMTU			= interfacePmtusize_match.group( 2 )

	def __repr__( self ):
		return "CfgNICmtu: mInterface: %s mMTU: %s" % (
				self.mInterface
			,	self.mMTU )

####################################################################################################

class vi_network_directive__AOE__OBJ(vi_network_directive__base):

	def GetDirectiveType( self ):
		return VI_NETWORK_DIRECTIVE__AOE

	def __init__( self, node_obj ):

		vi_network_directive__base.__init__( self )

		self.mInterface	= _INTERNAL__extract_expected_xml_text( node_obj, 'CfgNICaoe' ).strip()

	def __repr__( self ):
		return "CfgNICaoe: mInterface: %s" % ( self.mInterface )

####################################################################################################

class vi_network_directive__bond__OBJ(vi_network_directive__base):

	def GetDirectiveType( self ):
		return VI_NETWORK_DIRECTIVE__BOND

	def __init__( self, node_obj ):

		vi_network_directive__base.__init__( self )

		if node_obj.tagName != 'CfgBond':
			raise ValueError, "Given node_obj.tagName '%s' does NOT match expected 'CfgBond'" % ( node_obj.tagName )

		elemNode_LIST	= []
		for childNode in node_obj.childNodes:
			if childNode.nodeType == xml.dom.Node.ELEMENT_NODE:
				elemNode_LIST.append( childNode )
				continue
			if childNode.nodeType == xml.dom.Node.COMMENT_NODE:
				continue	# Ignore COMMENT_NODEs
			if childNode.nodeType == xml.dom.Node.TEXT_NODE:
				continue	# Ignore TEXT_NODE(whitespace)s
			raise ValueError, "Given CfgBond contains non-text node : %d : '%s'" % ( childNode.nodeType, childNode.toxml() )

		if len( elemNode_LIST ) < 4:
			raise ValueError, "Given CfgBond contains incorrect number of text nodes : %d" % ( len( elemNode_LIST ) )

		self.mBondTag				= _INTERNAL__extract_expected_xml_text( elemNode_LIST.pop(0), 'BondTag' ).strip()
		self.mBondType				= _INTERNAL__extract_expected_xml_text( elemNode_LIST.pop(0), 'BondType' ).strip()
		numOfInterfacesInBondStr	= _INTERNAL__extract_expected_xml_text( elemNode_LIST.pop(0), 'NumOfInterfacesInBond' ).strip()

		if int( numOfInterfacesInBondStr ) != len( elemNode_LIST ):
			raise ValueError, "Invalid NumOfInterfacesInBond %s vs %d" % ( numOfInterfacesInBondStr, len( elemNode_LIST ) )

		self.mInterface_LIST	= []
		for bondInterfaceNode in elemNode_LIST:
			bondInterface		= _INTERNAL__extract_expected_xml_text( bondInterfaceNode, 'Interface' ).strip()
			self.mInterface_LIST.append( bondInterface )

	def __repr__( self ):
		return "CfgBond: mBondTag: %s mBondType: %s mInterface_LIST: %s" % (
				self.mBondTag
			,	self.mBondType
			,	self.mInterface_LIST )

####################################################################################################

class vi_network_directive__route__OBJ(vi_network_directive__base):

	def GetDirectiveType( self ):
		return VI_NETWORK_DIRECTIVE__ROUTE

	def __init__( self, node_obj ):

		vi_network_directive__base.__init__( self )

		self.mRouteCmd	= _INTERNAL__extract_expected_xml_text( node_obj, 'CfgRoute' )

	def __repr__( self ):
		return "CfgRoute: mRouteCmd: '%s'" % ( self.mRouteCmd )

####################################################################################################

class vi_network_directive__ifconfig__OBJ(vi_network_directive__base):

	def GetDirectiveType( self ):
		return VI_NETWORK_DIRECTIVE__IFCONFIG

	def __init__( self, node_obj ):

		vi_network_directive__base.__init__( self )

		self.mIfconfigCmd	= _INTERNAL__extract_expected_xml_text( node_obj, 'CfgIfconfig' )

	def __repr__( self ):
		return "CfgIfconfig: mIfconfigCmd: '%s'" % ( self.mIfconfigCmd )

####################################################################################################

class vi_network_directive__ethtool__OBJ(vi_network_directive__base):

	def GetDirectiveType( self ):
		return VI_NETWORK_DIRECTIVE__ETHTOOL

	def __init__( self, node_obj ):

		vi_network_directive__base.__init__( self )

		self.mEthtoolCmd	= _INTERNAL__extract_expected_xml_text( node_obj, 'CfgEthtool' )

	def __repr__( self ):
		return "CfgEthtool: mEthtoolCmd: '%s'" % ( self.mEthtoolCmd )

####################################################################################################
def vi_get_default_network_directive_fcn_list():
	"""
	This routine
	"""
#
	def ret_None(*x):	return None

	networkCfgDirectiveFcn_LIST	= map( ret_None, range( VI_NETWORK_DIRECTIVE__range[1] ) )

	return networkCfgDirectiveFcn_LIST

####################################################################################################
def vi_extract_network_directive_list( vi_errorAndLog_obj, network_directives_xml_lines_buffer, networkCfgDirectiveFcn_LIST ):
	"""
	This routine parses the given network_directives_xml_lines_LIST into the list of
	directives (vi_network_directive__base subclasses) and returns them.
	"""
#
	vi_errorAndLog_obj.vi_print_dbg( 3, "XMLlines: '%s'" % ( network_directives_xml_lines_buffer ) )

	if len( networkCfgDirectiveFcn_LIST ) < VI_NETWORK_DIRECTIVE__range[1]:
		vi_errorAndLog_obj.vi_print_err( "Given networkCfgDirectiveFcn_LIST is too short : %d" % ( len( networkCfgDirectiveFcn_LIST ) ) )
		return []

	try:
		network_directives_dom	= xml.dom.minidom.parseString( network_directives_xml_lines_buffer )
	except:
		vi_errorAndLog_obj.vi_print_err( "Failed to parse network_directives_xml '%s' : \n%s" % ( network_directives_xml_lines_buffer, traceback.format_exc() ) )
		return []

	if not network_directives_dom.firstChild:
		vi_errorAndLog_obj.vi_print_wrn( "vi_extract_network_directive_list() Not even <NetworkCfgDirectives></NetworkCfgDirectives> specified" )
		return []

	network_directives_directive_LIST	= []

	for childNode in network_directives_dom.firstChild.childNodes:

		if childNode.nodeType != xml.dom.Node.ELEMENT_NODE:
			continue	# Ignore COMMENT_NODEs and TEXT_NODE(whitespace)s

		vi_errorAndLog_obj.vi_print_dbg( 3, "netWdirctV:", childNode.toxml() )

		newTask	= None
		try:
			attr_extraTimeUnicode	= childNode.getAttribute( 'extraTime' )
			if attr_extraTimeUnicode:
				attr_extraTime	= int(attr_extraTimeUnicode)
				if attr_extraTime < 0:
					vi_errorAndLog_obj.vi_print_wrn( "extraTime attr for '%s' was negative : %d" % ( childNode.toxml(), attr_extraTime ) )
					attr_extraTime	= 0
			else:
				attr_extraTime	= None

			if   childNode.tagName == 'CfgNICdhcp':
				newTask		= vi_network_directive__DHCP__OBJ( childNode )
			elif childNode.tagName == 'CfgNICstatic':
				newTask		= vi_network_directive__static_IP__OBJ( childNode )
			elif childNode.tagName == 'CfgNICmtu':
				newTask		= vi_network_directive__MTU__OBJ( childNode )
			elif childNode.tagName == 'CfgNICaoe':
				newTask		= vi_network_directive__AOE__OBJ( childNode )
			elif childNode.tagName == 'CfgBond':
				newTask		= vi_network_directive__bond__OBJ( childNode )
			elif childNode.tagName == 'CfgRoute':
				newTask		= vi_network_directive__route__OBJ( childNode )
			elif childNode.tagName == 'CfgIfconfig':
				newTask		= vi_network_directive__ifconfig__OBJ( childNode )
			elif childNode.tagName == 'CfgEthtool':
				newTask		= vi_network_directive__ethtool__OBJ( childNode )
			else:
				vi_errorAndLog_obj.vi_print_wrn( "Unknown NetworkCfgDirectives element node.tagName '%s'" % ( childNode.tagName ) )
		except:
			vi_errorAndLog_obj.vi_print_err( "Failed to create directive for '%s' : '%s' : \n%s" % ( childNode.tagName, childNode.toxml(), traceback.format_exc() ) )

		if newTask:
			if attr_extraTime:
				newTask.mAttr_extraTime	= attr_extraTime
			newTask.mFcn	= networkCfgDirectiveFcn_LIST[ newTask.GetDirectiveType() ]
			network_directives_directive_LIST.append( newTask )

	network_directives_dom.unlink()	# This MUST be done or the memory will NOT be reclaimed !!!

	return network_directives_directive_LIST

####################################################################################################
# END
